@@ -19,26 +19,36 @@ abstract class EfficientLengthIterable<T> extends Iterable<T> {
   int get length;
 }
 
+/// An interface which hides [EfficientLengthIterable] from upper bounds.
+///
+/// Every type which implements [EfficientLengthIterable] also implements
+/// this interface, and they have the same *depth*, so it's impossible
+/// for the upper-bound algorithm to get [EfficientLengthIterable]
+/// as the result.
+abstract interface class HideEfficientLengthIterable<T>
+    implements Iterable<T> {}
+
 /**
  * An [Iterable] for classes that have efficient [length] and [elementAt].
  *
  * All other methods are implemented in terms of [length] and [elementAt],
  * including [iterator].
  */
-abstract class ListIterable<E> extends EfficientLengthIterable<E> {
+abstract class ListIterable<E> extends EfficientLengthIterable<E>
+    implements HideEfficientLengthIterable<E> {
   int get length;
   E elementAt(int i);
 
   const ListIterable();
 
-  Iterator<E> get iterator => new ListIterator<E>(this);
+  Iterator<E> get iterator => ListIterator<E>(this);
 
   void forEach(void action(E element)) {
     int length = this.length;
     for (int i = 0; i < length; i++) {
       action(elementAt(i));
       if (length != this.length) {
-        throw new ConcurrentModificationError(this);
+        throw ConcurrentModificationError(this);
       }
     }
   }
@@ -61,12 +71,12 @@ abstract class ListIterable<E> extends EfficientLengthIterable<E> {
     return elementAt(0);
   }
 
-  bool contains(Object element) {
+  bool contains(Object? element) {
     int length = this.length;
     for (int i = 0; i < length; i++) {
       if (elementAt(i) == element) return true;
       if (length != this.length) {
-        throw new ConcurrentModificationError(this);
+        throw ConcurrentModificationError(this);
       }
     }
     return false;
@@ -77,7 +87,7 @@ abstract class ListIterable<E> extends EfficientLengthIterable<E> {
     for (int i = 0; i < length; i++) {
       if (!test(elementAt(i))) return false;
       if (length != this.length) {
-        throw new ConcurrentModificationError(this);
+        throw ConcurrentModificationError(this);
       }
     }
     return true;
@@ -88,41 +98,41 @@ abstract class ListIterable<E> extends EfficientLengthIterable<E> {
     for (int i = 0; i < length; i++) {
       if (test(elementAt(i))) return true;
       if (length != this.length) {
-        throw new ConcurrentModificationError(this);
+        throw ConcurrentModificationError(this);
       }
     }
     return false;
   }
 
-  E firstWhere(bool test(E element), {E orElse()}) {
+  E firstWhere(bool test(E element), {E Function()? orElse}) {
     int length = this.length;
     for (int i = 0; i < length; i++) {
       E element = elementAt(i);
       if (test(element)) return element;
       if (length != this.length) {
-        throw new ConcurrentModificationError(this);
+        throw ConcurrentModificationError(this);
       }
     }
     if (orElse != null) return orElse();
     throw IterableElementError.noElement();
   }
 
-  E lastWhere(bool test(E element), {E orElse()}) {
+  E lastWhere(bool test(E element), {E Function()? orElse}) {
     int length = this.length;
     for (int i = length - 1; i >= 0; i--) {
       E element = elementAt(i);
       if (test(element)) return element;
       if (length != this.length) {
-        throw new ConcurrentModificationError(this);
+        throw ConcurrentModificationError(this);
       }
     }
     if (orElse != null) return orElse();
     throw IterableElementError.noElement();
   }
 
-  E singleWhere(bool test(E element), {E orElse()}) {
+  E singleWhere(bool test(E element), {E Function()? orElse}) {
     int length = this.length;
-    E match = null;
+    late E match;
     bool matchFound = false;
     for (int i = 0; i < length; i++) {
       E element = elementAt(i);
@@ -134,7 +144,7 @@ abstract class ListIterable<E> extends EfficientLengthIterable<E> {
         match = element;
       }
       if (length != this.length) {
-        throw new ConcurrentModificationError(this);
+        throw ConcurrentModificationError(this);
       }
     }
     if (matchFound) return match;
@@ -148,23 +158,23 @@ abstract class ListIterable<E> extends EfficientLengthIterable<E> {
       if (length == 0) return "";
       String first = "${elementAt(0)}";
       if (length != this.length) {
-        throw new ConcurrentModificationError(this);
+        throw ConcurrentModificationError(this);
       }
-      StringBuffer buffer = new StringBuffer(first);
+      StringBuffer buffer = StringBuffer(first);
       for (int i = 1; i < length; i++) {
         buffer.write(separator);
         buffer.write(elementAt(i));
         if (length != this.length) {
-          throw new ConcurrentModificationError(this);
+          throw ConcurrentModificationError(this);
         }
       }
       return buffer.toString();
     } else {
-      StringBuffer buffer = new StringBuffer();
+      StringBuffer buffer = StringBuffer();
       for (int i = 0; i < length; i++) {
         buffer.write(elementAt(i));
         if (length != this.length) {
-          throw new ConcurrentModificationError(this);
+          throw ConcurrentModificationError(this);
         }
       }
       return buffer.toString();
@@ -173,7 +183,8 @@ abstract class ListIterable<E> extends EfficientLengthIterable<E> {
 
   Iterable<E> where(bool test(E element)) => super.where(test);
 
-  Iterable<T> map<T>(T f(E element)) => new MappedListIterable<E, T>(this, f);
+  Iterable<T> map<T>(T toElement(E element)) =>
+      MappedListIterable<E, T>(this, toElement);
 
   E reduce(E combine(E value, E element)) {
     int length = this.length;
@@ -182,7 +193,7 @@ abstract class ListIterable<E> extends EfficientLengthIterable<E> {
     for (int i = 1; i < length; i++) {
       value = combine(value, elementAt(i));
       if (length != this.length) {
-        throw new ConcurrentModificationError(this);
+        throw ConcurrentModificationError(this);
       }
     }
     return value;
@@ -194,35 +205,26 @@ abstract class ListIterable<E> extends EfficientLengthIterable<E> {
     for (int i = 0; i < length; i++) {
       value = combine(value, elementAt(i));
       if (length != this.length) {
-        throw new ConcurrentModificationError(this);
+        throw ConcurrentModificationError(this);
       }
     }
     return value;
   }
 
-  Iterable<E> skip(int count) => new SubListIterable<E>(this, count, null);
+  Iterable<E> skip(int count) => SubListIterable<E>(this, count, null);
 
   Iterable<E> skipWhile(bool test(E element)) => super.skipWhile(test);
 
-  Iterable<E> take(int count) => new SubListIterable<E>(this, 0, count);
+  Iterable<E> take(int count) =>
+      SubListIterable<E>(this, 0, checkNotNullable(count, "count"));
 
   Iterable<E> takeWhile(bool test(E element)) => super.takeWhile(test);
 
-  List<E> toList({bool growable: true}) {
-    List<E> result;
-    if (growable) {
-      result = <E>[]..length = length;
-    } else {
-      result = new List<E>(length);
-    }
-    for (int i = 0; i < length; i++) {
-      result[i] = elementAt(i);
-    }
-    return result;
-  }
+  List<E> toList({bool growable = true}) =>
+      List<E>.of(this, growable: growable);
 
   Set<E> toSet() {
-    Set<E> result = new Set<E>();
+    Set<E> result = Set<E>();
     for (int i = 0; i < length; i++) {
       result.add(elementAt(i));
     }
@@ -230,26 +232,28 @@ abstract class ListIterable<E> extends EfficientLengthIterable<E> {
   }
 }
 
-class SubListIterable<E> extends ListIterable<E> {
+base class SubListIterable<E> extends ListIterable<E> {
   final Iterable<E> _iterable; // Has efficient length and elementAt.
   final int _start;
   /** If null, represents the length of the iterable. */
-  final int _endOrLength;
+  final int? _endOrLength;
 
   SubListIterable(this._iterable, this._start, this._endOrLength) {
     RangeError.checkNotNegative(_start, "start");
-    if (_endOrLength != null) {
-      RangeError.checkNotNegative(_endOrLength, "end");
-      if (_start > _endOrLength) {
-        throw new RangeError.range(_start, 0, _endOrLength, "start");
+    int? endOrLength = _endOrLength;
+    if (endOrLength != null) {
+      RangeError.checkNotNegative(endOrLength, "end");
+      if (_start > endOrLength) {
+        throw RangeError.range(_start, 0, endOrLength, "start");
       }
     }
   }
 
   int get _endIndex {
     int length = _iterable.length;
-    if (_endOrLength == null || _endOrLength > length) return length;
-    return _endOrLength;
+    int? endOrLength = _endOrLength;
+    if (endOrLength == null || endOrLength > length) return length;
+    return endOrLength;
   }
 
   int get _startIndex {
@@ -261,16 +265,18 @@ class SubListIterable<E> extends ListIterable<E> {
   int get length {
     int length = _iterable.length;
     if (_start >= length) return 0;
-    if (_endOrLength == null || _endOrLength >= length) {
+    int? endOrLength = _endOrLength;
+    if (endOrLength == null || endOrLength >= length) {
       return length - _start;
     }
-    return _endOrLength - _start;
+    return endOrLength - _start;
   }
 
   E elementAt(int index) {
     int realIndex = _startIndex + index;
     if (index < 0 || realIndex >= _endIndex) {
-      throw new RangeError.index(index, this, "index");
+      throw IndexError.withLength(index, length,
+          indexable: this, name: "index");
     }
     return _iterable.elementAt(realIndex);
   }
@@ -278,34 +284,38 @@ class SubListIterable<E> extends ListIterable<E> {
   Iterable<E> skip(int count) {
     RangeError.checkNotNegative(count, "count");
     int newStart = _start + count;
-    if (_endOrLength != null && newStart >= _endOrLength) {
-      return new EmptyIterable<E>();
+    int? endOrLength = _endOrLength;
+    if (endOrLength != null && newStart >= endOrLength) {
+      return EmptyIterable<E>();
     }
-    return new SubListIterable<E>(_iterable, newStart, _endOrLength);
+    return SubListIterable<E>(_iterable, newStart, _endOrLength);
   }
 
   Iterable<E> take(int count) {
     RangeError.checkNotNegative(count, "count");
-    if (_endOrLength == null) {
-      return new SubListIterable<E>(_iterable, _start, _start + count);
+    int? endOrLength = _endOrLength;
+    if (endOrLength == null) {
+      return SubListIterable<E>(_iterable, _start, _start + count);
     } else {
       int newEnd = _start + count;
-      if (_endOrLength < newEnd) return this;
-      return new SubListIterable<E>(_iterable, _start, newEnd);
+      if (endOrLength < newEnd) return this;
+      return SubListIterable<E>(_iterable, _start, newEnd);
     }
   }
 
-  List<E> toList({bool growable: true}) {
+  List<E> toList({bool growable = true}) {
     int start = _start;
     int end = _iterable.length;
-    if (_endOrLength != null && _endOrLength < end) end = _endOrLength;
+    int? endOrLength = _endOrLength;
+    if (endOrLength != null && endOrLength < end) end = endOrLength;
     int length = end - start;
-    if (length < 0) length = 0;
+    if (length <= 0) return List<E>.empty(growable: growable);
+
     List<E> result =
-        growable ? (new List<E>()..length = length) : new List<E>(length);
-    for (int i = 0; i < length; i++) {
+        List<E>.filled(length, _iterable.elementAt(start), growable: growable);
+    for (int i = 1; i < length; i++) {
       result[i] = _iterable.elementAt(start + i);
-      if (_iterable.length < end) throw new ConcurrentModificationError(this);
+      if (_iterable.length < end) throw ConcurrentModificationError(this);
     }
     return result;
   }
@@ -322,19 +332,20 @@ class ListIterator<E> implements Iterator<E> {
   final Iterable<E> _iterable;
   final int _length;
   int _index;
-  E _current;
+  E? _current;
 
   ListIterator(Iterable<E> iterable)
       : _iterable = iterable,
         _length = iterable.length,
         _index = 0;
 
-  E get current => _current;
+  E get current => _current as E;
 
+  @pragma("vm:prefer-inline")
   bool moveNext() {
     int length = _iterable.length;
     if (_length != length) {
-      throw new ConcurrentModificationError(_iterable);
+      throw ConcurrentModificationError(_iterable);
     }
     if (_index >= length) {
       _current = null;
@@ -354,14 +365,14 @@ class MappedIterable<S, T> extends Iterable<T> {
 
   factory MappedIterable(Iterable<S> iterable, T function(S value)) {
     if (iterable is EfficientLengthIterable) {
-      return new EfficientLengthMappedIterable<S, T>(iterable, function);
+      return EfficientLengthMappedIterable<S, T>(iterable, function);
     }
-    return new MappedIterable<S, T>._(iterable, function);
+    return MappedIterable<S, T>._(iterable, function);
   }
 
   MappedIterable._(this._iterable, this._f);
 
-  Iterator<T> get iterator => new MappedIterator<S, T>(_iterable.iterator, _f);
+  Iterator<T> get iterator => MappedIterator<S, T>(_iterable.iterator, _f);
 
   // Length related functions are independent of the mapping.
   int get length => _iterable.length;
@@ -375,13 +386,13 @@ class MappedIterable<S, T> extends Iterable<T> {
 }
 
 class EfficientLengthMappedIterable<S, T> extends MappedIterable<S, T>
-    implements EfficientLengthIterable<T> {
+    implements EfficientLengthIterable<T>, HideEfficientLengthIterable<T> {
   EfficientLengthMappedIterable(Iterable<S> iterable, T function(S value))
       : super._(iterable, function);
 }
 
-class MappedIterator<S, T> extends Iterator<T> {
-  T _current;
+class MappedIterator<S, T> implements Iterator<T> {
+  T? _current;
   final Iterator<S> _iterator;
   final _Transformation<S, T> _f;
 
@@ -396,7 +407,7 @@ class MappedIterator<S, T> extends Iterator<T> {
     return false;
   }
 
-  T get current => _current;
+  T get current => _current as T;
 }
 
 /**
@@ -404,7 +415,7 @@ class MappedIterator<S, T> extends Iterator<T> {
  *
  * Expects efficient `length` and `elementAt` on the source iterable.
  */
-class MappedListIterable<S, T> extends ListIterable<T> {
+base class MappedListIterable<S, T> extends ListIterable<T> {
   final Iterable<S> _source;
   final _Transformation<S, T> _f;
 
@@ -422,13 +433,14 @@ class WhereIterable<E> extends Iterable<E> {
 
   WhereIterable(this._iterable, this._f);
 
-  Iterator<E> get iterator => new WhereIterator<E>(_iterable.iterator, _f);
+  Iterator<E> get iterator => WhereIterator<E>(_iterable.iterator, _f);
 
   // Specialization of [Iterable.map] to non-EfficientLengthIterable.
-  Iterable<T> map<T>(T f(E element)) => new MappedIterable<E, T>._(this, f);
+  Iterable<T> map<T>(T toElement(E element)) =>
+      MappedIterable<E, T>._(this, toElement);
 }
 
-class WhereIterator<E> extends Iterator<E> {
+class WhereIterator<E> implements Iterator<E> {
   final Iterator<E> _iterator;
   final _ElementPredicate<E> _f;
 
@@ -454,7 +466,7 @@ class ExpandIterable<S, T> extends Iterable<T> {
 
   ExpandIterable(this._iterable, this._f);
 
-  Iterator<T> get iterator => new ExpandIterator<S, T>(_iterable.iterator, _f);
+  Iterator<T> get iterator => ExpandIterator<S, T>(_iterable.iterator, _f);
 }
 
 class ExpandIterator<S, T> implements Iterator<T> {
@@ -463,16 +475,16 @@ class ExpandIterator<S, T> implements Iterator<T> {
   // Initialize _currentExpansion to an empty iterable. A null value
   // marks the end of iteration, and we don't want to call _f before
   // the first moveNext call.
-  Iterator<T> _currentExpansion = const EmptyIterator();
-  T _current;
+  Iterator<T>? _currentExpansion = const EmptyIterator<Never>();
+  T? _current;
 
   ExpandIterator(this._iterator, this._f);
 
-  T get current => _current;
+  T get current => _current as T;
 
   bool moveNext() {
     if (_currentExpansion == null) return false;
-    while (!_currentExpansion.moveNext()) {
+    while (!_currentExpansion!.moveNext()) {
       _current = null;
       if (_iterator.moveNext()) {
         // If _f throws, this ends iteration. Otherwise _currentExpansion and
@@ -483,7 +495,7 @@ class ExpandIterator<S, T> implements Iterator<T> {
         return false;
       }
     }
-    _current = _currentExpansion.current;
+    _current = _currentExpansion!.current;
     return true;
   }
 }
@@ -493,24 +505,23 @@ class TakeIterable<E> extends Iterable<E> {
   final int _takeCount;
 
   factory TakeIterable(Iterable<E> iterable, int takeCount) {
-    if (takeCount is! int || takeCount < 0) {
-      throw new ArgumentError(takeCount);
-    }
+    ArgumentError.checkNotNull(takeCount, "takeCount");
+    RangeError.checkNotNegative(takeCount, "takeCount");
     if (iterable is EfficientLengthIterable) {
-      return new EfficientLengthTakeIterable<E>(iterable, takeCount);
+      return EfficientLengthTakeIterable<E>(iterable, takeCount);
     }
-    return new TakeIterable<E>._(iterable, takeCount);
+    return TakeIterable<E>._(iterable, takeCount);
   }
 
   TakeIterable._(this._iterable, this._takeCount);
 
   Iterator<E> get iterator {
-    return new TakeIterator<E>(_iterable.iterator, _takeCount);
+    return TakeIterator<E>(_iterable.iterator, _takeCount);
   }
 }
 
 class EfficientLengthTakeIterable<E> extends TakeIterable<E>
-    implements EfficientLengthIterable<E> {
+    implements EfficientLengthIterable<E>, HideEfficientLengthIterable<E> {
   EfficientLengthTakeIterable(Iterable<E> iterable, int takeCount)
       : super._(iterable, takeCount);
 
@@ -521,12 +532,12 @@ class EfficientLengthTakeIterable<E> extends TakeIterable<E>
   }
 }
 
-class TakeIterator<E> extends Iterator<E> {
+class TakeIterator<E> implements Iterator<E> {
   final Iterator<E> _iterator;
   int _remaining;
 
   TakeIterator(this._iterator, this._remaining) {
-    assert(_remaining is int && _remaining >= 0);
+    assert(_remaining >= 0);
   }
 
   bool moveNext() {
@@ -539,7 +550,11 @@ class TakeIterator<E> extends Iterator<E> {
   }
 
   E get current {
-    if (_remaining < 0) return null;
+    // Before NNBD, this returned null when iteration was complete. In order to
+    // avoid a hard breaking change, we return "null as E" in that case so that
+    // if strong checking is not enabled or E is nullable, the existing
+    // behavior is preserved.
+    if (_remaining < 0) return null as E;
     return _iterator.current;
   }
 }
@@ -551,11 +566,11 @@ class TakeWhileIterable<E> extends Iterable<E> {
   TakeWhileIterable(this._iterable, this._f);
 
   Iterator<E> get iterator {
-    return new TakeWhileIterator<E>(_iterable.iterator, _f);
+    return TakeWhileIterator<E>(_iterable.iterator, _f);
   }
 }
 
-class TakeWhileIterator<E> extends Iterator<E> {
+class TakeWhileIterator<E> implements Iterator<E> {
   final Iterator<E> _iterator;
   final _ElementPredicate<E> _f;
   bool _isFinished = false;
@@ -572,7 +587,7 @@ class TakeWhileIterator<E> extends Iterator<E> {
   }
 
   E get current {
-    if (_isFinished) return null;
+    if (_isFinished) return null as E;
     return _iterator.current;
   }
 }
@@ -583,26 +598,26 @@ class SkipIterable<E> extends Iterable<E> {
 
   factory SkipIterable(Iterable<E> iterable, int count) {
     if (iterable is EfficientLengthIterable) {
-      return new EfficientLengthSkipIterable<E>(iterable, count);
+      return EfficientLengthSkipIterable<E>(iterable, count);
     }
-    return new SkipIterable<E>._(iterable, _checkCount(count));
+    return SkipIterable<E>._(iterable, _checkCount(count));
   }
 
   SkipIterable._(this._iterable, this._skipCount);
 
   Iterable<E> skip(int count) {
-    return new SkipIterable<E>._(_iterable, _skipCount + _checkCount(count));
+    return SkipIterable<E>._(_iterable, _skipCount + _checkCount(count));
   }
 
   Iterator<E> get iterator {
-    return new SkipIterator<E>(_iterable.iterator, _skipCount);
+    return SkipIterator<E>(_iterable.iterator, _skipCount);
   }
 }
 
 class EfficientLengthSkipIterable<E> extends SkipIterable<E>
-    implements EfficientLengthIterable<E> {
+    implements EfficientLengthIterable<E>, HideEfficientLengthIterable<E> {
   factory EfficientLengthSkipIterable(Iterable<E> iterable, int count) {
-    return new EfficientLengthSkipIterable<E>._(iterable, _checkCount(count));
+    return EfficientLengthSkipIterable<E>._(iterable, _checkCount(count));
   }
 
   EfficientLengthSkipIterable._(Iterable<E> iterable, int count)
@@ -615,25 +630,23 @@ class EfficientLengthSkipIterable<E> extends SkipIterable<E>
   }
 
   Iterable<E> skip(int count) {
-    return new EfficientLengthSkipIterable<E>._(
+    return EfficientLengthSkipIterable<E>._(
         _iterable, _skipCount + _checkCount(count));
   }
 }
 
 int _checkCount(int count) {
-  if (count is! int) {
-    throw new ArgumentError.value(count, "count", "is not an integer");
-  }
+  ArgumentError.checkNotNull(count, "count");
   RangeError.checkNotNegative(count, "count");
   return count;
 }
 
-class SkipIterator<E> extends Iterator<E> {
+class SkipIterator<E> implements Iterator<E> {
   final Iterator<E> _iterator;
   int _skipCount;
 
   SkipIterator(this._iterator, this._skipCount) {
-    assert(_skipCount is int && _skipCount >= 0);
+    assert(_skipCount >= 0);
   }
 
   bool moveNext() {
@@ -652,11 +665,11 @@ class SkipWhileIterable<E> extends Iterable<E> {
   SkipWhileIterable(this._iterable, this._f);
 
   Iterator<E> get iterator {
-    return new SkipWhileIterator<E>(_iterable.iterator, _f);
+    return SkipWhileIterator<E>(_iterable.iterator, _f);
   }
 }
 
-class SkipWhileIterator<E> extends Iterator<E> {
+class SkipWhileIterator<E> implements Iterator<E> {
   final Iterator<E> _iterator;
   final _ElementPredicate<E> _f;
   bool _hasSkipped = false;
@@ -679,10 +692,11 @@ class SkipWhileIterator<E> extends Iterator<E> {
 /**
  * The always empty [Iterable].
  */
-class EmptyIterable<E> extends EfficientLengthIterable<E> {
+class EmptyIterable<E> extends EfficientLengthIterable<E>
+    implements HideEfficientLengthIterable<E> {
   const EmptyIterable();
 
-  Iterator<E> get iterator => const EmptyIterator();
+  Iterator<E> get iterator => const EmptyIterator<Never>();
 
   void forEach(void action(E element)) {}
 
@@ -703,26 +717,26 @@ class EmptyIterable<E> extends EfficientLengthIterable<E> {
   }
 
   E elementAt(int index) {
-    throw new RangeError.range(index, 0, 0, "index");
+    throw RangeError.range(index, 0, 0, "index");
   }
 
-  bool contains(Object element) => false;
+  bool contains(Object? element) => false;
 
   bool every(bool test(E element)) => true;
 
   bool any(bool test(E element)) => false;
 
-  E firstWhere(bool test(E element), {E orElse()}) {
+  E firstWhere(bool test(E element), {E Function()? orElse}) {
     if (orElse != null) return orElse();
     throw IterableElementError.noElement();
   }
 
-  E lastWhere(bool test(E element), {E orElse()}) {
+  E lastWhere(bool test(E element), {E Function()? orElse}) {
     if (orElse != null) return orElse();
     throw IterableElementError.noElement();
   }
 
-  E singleWhere(bool test(E element), {E orElse()}) {
+  E singleWhere(bool test(E element), {E Function()? orElse}) {
     if (orElse != null) return orElse();
     throw IterableElementError.noElement();
   }
@@ -731,7 +745,7 @@ class EmptyIterable<E> extends EfficientLengthIterable<E> {
 
   Iterable<E> where(bool test(E element)) => this;
 
-  Iterable<T> map<T>(T f(E element)) => new EmptyIterable<T>();
+  Iterable<T> map<T>(T toElement(E element)) => EmptyIterable<T>();
 
   E reduce(E combine(E value, E element)) {
     throw IterableElementError.noElement();
@@ -755,16 +769,18 @@ class EmptyIterable<E> extends EfficientLengthIterable<E> {
 
   Iterable<E> takeWhile(bool test(E element)) => this;
 
-  List<E> toList({bool growable: true}) => growable ? <E>[] : new List<E>(0);
+  List<E> toList({bool growable = true}) => List<E>.empty(growable: growable);
 
-  Set<E> toSet() => new Set<E>();
+  Set<E> toSet() => Set<E>();
 }
 
 /** The always empty iterator. */
 class EmptyIterator<E> implements Iterator<E> {
   const EmptyIterator();
   bool moveNext() => false;
-  E get current => null;
+  E get current {
+    throw IterableElementError.noElement();
+  }
 }
 
 class FollowedByIterable<E> extends Iterable<E> {
@@ -775,19 +791,19 @@ class FollowedByIterable<E> extends Iterable<E> {
   factory FollowedByIterable.firstEfficient(
       EfficientLengthIterable<E> first, Iterable<E> second) {
     if (second is EfficientLengthIterable<E>) {
-      return new EfficientLengthFollowedByIterable<E>(first, second);
+      return EfficientLengthFollowedByIterable<E>(first, second);
     }
-    return new FollowedByIterable<E>(first, second);
+    return FollowedByIterable<E>(first, second);
   }
 
-  Iterator<E> get iterator => new FollowedByIterator(_first, _second);
+  Iterator<E> get iterator => FollowedByIterator(_first, _second);
 
   int get length => _first.length + _second.length;
   bool get isEmpty => _first.isEmpty && _second.isEmpty;
   bool get isNotEmpty => _first.isNotEmpty || _second.isNotEmpty;
 
   // May be more efficient if either iterable is a Set.
-  bool contains(Object value) =>
+  bool contains(Object? value) =>
       _first.contains(value) || _second.contains(value);
 
   E get first {
@@ -811,25 +827,11 @@ class FollowedByIterable<E> extends Iterable<E> {
 }
 
 class EfficientLengthFollowedByIterable<E> extends FollowedByIterable<E>
-    implements EfficientLengthIterable<E> {
+    implements EfficientLengthIterable<E>, HideEfficientLengthIterable<E> {
   EfficientLengthFollowedByIterable(
       EfficientLengthIterable<E> first, EfficientLengthIterable<E> second)
       : super(first, second);
 
-  Iterable<E> skip(int count) {
-    int firstLength = _first.length;
-    if (count >= firstLength) return _second.skip(count - firstLength);
-    return new EfficientLengthFollowedByIterable<E>(
-        _first.skip(count), _second);
-  }
-
-  Iterable<E> take(int count) {
-    int firstLength = _first.length;
-    if (count <= firstLength) return _first.take(count);
-    return new EfficientLengthFollowedByIterable<E>(
-        _first, _second.take(count - firstLength));
-  }
-
   E elementAt(int index) {
     int firstLength = _first.length;
     if (index < firstLength) return _first.elementAt(index);
@@ -849,7 +851,7 @@ class EfficientLengthFollowedByIterable<E> extends FollowedByIterable<E>
 
 class FollowedByIterator<E> implements Iterator<E> {
   Iterator<E> _currentIterator;
-  Iterable<E> _nextIterable;
+  Iterable<E>? _nextIterable;
 
   FollowedByIterator(Iterable<E> first, this._nextIterable)
       : _currentIterator = first.iterator;
@@ -857,7 +859,7 @@ class FollowedByIterator<E> implements Iterator<E> {
   bool moveNext() {
     if (_currentIterator.moveNext()) return true;
     if (_nextIterable != null) {
-      _currentIterator = _nextIterable.iterator;
+      _currentIterator = _nextIterable!.iterator;
       _nextIterable = null;
       return _currentIterator.moveNext();
     }
@@ -868,13 +870,13 @@ class FollowedByIterator<E> implements Iterator<E> {
 }
 
 class WhereTypeIterable<T> extends Iterable<T> {
-  final Iterable<Object> _source;
+  final Iterable<Object?> _source;
   WhereTypeIterable(this._source);
-  Iterator<T> get iterator => new WhereTypeIterator<T>(_source.iterator);
+  Iterator<T> get iterator => WhereTypeIterator<T>(_source.iterator);
 }
 
 class WhereTypeIterator<T> implements Iterator<T> {
-  final Iterator<Object> _source;
+  final Iterator<Object?> _source;
   WhereTypeIterator(this._source);
   bool moveNext() {
     while (_source.moveNext()) {
@@ -883,17 +885,170 @@ class WhereTypeIterator<T> implements Iterator<T> {
     return false;
   }
 
-  T get current => _source.current;
+  T get current => _source.current as T;
+}
+
+/// Implementation of [NullableIterableExtensions.nonNulls].
+///
+/// A filtering iterable, so it doesn't have efficient length
+/// and cannot forward most methods to the underlying [_source].
+class NonNullsIterable<T extends Object> extends Iterable<T> {
+  final Iterable<T?> _source;
+  NonNullsIterable(this._source);
+
+  T? get _firstNonNull {
+    for (var element in _source) {
+      if (element != null) return element;
+    }
+    return null;
+  }
+
+  bool get isEmpty => _firstNonNull == null;
+  bool get isNotEmpty => _firstNonNull != null;
+  T get first => _firstNonNull ?? (throw IterableElementError.noElement());
+
+  Iterator<T> get iterator => NonNullsIterator<T>(_source.iterator);
+}
+
+class NonNullsIterator<T extends Object> implements Iterator<T> {
+  final Iterator<T?> _source;
+  T? _current;
+
+  NonNullsIterator(this._source);
+
+  bool moveNext() {
+    _current = null;
+    while (_source.moveNext()) {
+      var next = _source.current;
+      if (next != null) {
+        _current = next;
+        return true;
+      }
+    }
+    return false;
+  }
+
+  T get current => _current ?? (throw IterableElementError.noElement());
+}
+
+/// Implementation of [IterableExtensions.indexed].
+///
+/// Maps elements of [_source] one-to-one to record values,
+/// so has the same length as the original, and can define many
+/// operations in terms of the underlying source.
+class IndexedIterable<T> extends Iterable<(int, T)> {
+  final Iterable<T> _source;
+
+  /// Offset applied to indices.
+  ///
+  /// Used to implement `skip` efficiently for iterables which can skip
+  /// efficiently.
+  final int _start;
+
+  @pragma('vm:prefer-inline')
+  factory IndexedIterable(Iterable<T> source, int start) {
+    if (source is EfficientLengthIterable) {
+      return EfficientLengthIndexedIterable(source, start);
+    }
+    return IndexedIterable._(source, start);
+  }
+
+  IndexedIterable.nonEfficientLength(Iterable<T> source, int start)
+      : this._(source, start);
+
+  IndexedIterable._(this._source, this._start);
+
+  int get length => _source.length;
+  bool get isEmpty => _source.isEmpty;
+  bool get isNotEmpty => _source.isNotEmpty;
+
+  (int, T) get first => (_start, _source.first);
+  (int, T) get single => (_start, _source.single);
+  (int, T) elementAt(int index) => (index + _start, _source.elementAt(index));
+
+  bool contains(Object? element) {
+    if (element case (int index, Object? other) when index >= _start) {
+      // Try to find the `index`th element without looking at the
+      // intermediate values, and without throwing if there are fewer.
+      var unbiasedIndex = index - _start;
+      var iterator = _source.skip(unbiasedIndex).iterator;
+      return iterator.moveNext() && iterator.current == other;
+    }
+    return false;
+  }
+
+  Iterable<(int, T)> take(int count) => IndexedIterable<T>.nonEfficientLength(
+      _source.take(_checkCount(count)), _start);
+
+  Iterable<(int, T)> skip(int count) => IndexedIterable<T>.nonEfficientLength(
+      _source.skip(_checkCount(count)), count + _start);
+
+  @pragma('vm:prefer-inline')
+  Iterator<(int, T)> get iterator =>
+      IndexedIterator<T>(_source.iterator, _start);
+}
+
+class EfficientLengthIndexedIterable<T> extends IndexedIterable<T>
+    implements
+        EfficientLengthIterable<(int, T)>,
+        HideEfficientLengthIterable<(int, T)> {
+  EfficientLengthIndexedIterable(super._source, super._start) : super._();
+
+  (int, T) get last {
+    var length = _source.length;
+    if (length <= 0) throw IterableElementError.noElement();
+    var last = _source.last;
+    if (length != this.length) {
+      throw ConcurrentModificationError(this);
+    }
+    return (length - 1 + _start, last);
+  }
+
+  bool contains(Object? element) {
+    if (element case (int index, Object? other) when index >= _start) {
+      var unbiasedIndex = index - _start;
+      return unbiasedIndex < _source.length &&
+          _source.elementAt(unbiasedIndex) == other;
+    }
+    return false;
+  }
+
+  Iterable<(int, T)> take(int count) => EfficientLengthIndexedIterable<T>(
+      _source.take(_checkCount(count)), _start);
+
+  Iterable<(int, T)> skip(int count) => EfficientLengthIndexedIterable<T>(
+      _source.skip(_checkCount(count)), _start + count);
+}
+
+class IndexedIterator<T> implements Iterator<(int, T)> {
+  final Iterator<T> _source;
+  final int _start;
+  int _index = -1;
+
+  IndexedIterator(this._source, this._start);
+
+  bool moveNext() {
+    var index = ++_index;
+    if (index >= 0 && _source.moveNext()) {
+      return true;
+    }
+    _index = -2; // Ensures moveNext won't get called again.
+    return false;
+  }
+
+  (int, T) get current => _index >= 0
+      ? (_start + _index, _source.current)
+      : (throw IterableElementError.noElement());
 }
 
 /**
  * Creates errors throw by [Iterable] when the element count is wrong.
  */
 abstract class IterableElementError {
-  /** Error thrown thrown by, e.g., [Iterable.first] when there is no result. */
-  static StateError noElement() => new StateError("No element");
+  /** Error thrown by, e.g., [Iterable.first] when there is no result. */
+  static StateError noElement() => StateError("No element");
   /** Error thrown by, e.g., [Iterable.single] if there are too many results. */
-  static StateError tooMany() => new StateError("Too many elements");
+  static StateError tooMany() => StateError("Too many elements");
   /** Error thrown by, e.g., [List.setRange] if there are too few elements. */
-  static StateError tooFew() => new StateError("Too few elements");
+  static StateError tooFew() => StateError("Too few elements");
 }
